/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2005 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#if !defined(AFX_BUTTONTEXTURE_H__4F10FDCD_EF77_423F_8223_011704DFFED2__INCLUDED_)
#define AFX_BUTTONTEXTURE_H__4F10FDCD_EF77_423F_8223_011704DFFED2__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

typedef vector<COLORREF> ColorRefList;
typedef vector<COLORREF>::iterator ColorRefListIterator;

struct BUTTONSETTINGS
{
	int borderMethod;

	int leftEdge;
	int topEdge;
	int rightEdge;
	int bottomEdge;

	int mode;

	int button_saturation;
	int button_hueIntensity;
	COLORREF button_hueColor;

	bool trueTransparent;
};

class ButtonTexture  
{
public:
	ButtonTexture();
	virtual ~ButtonTexture();

	void configure(const string &uniquekey, const string &groupkey);
	void apply(HDC hDC, int x, int y, int width, int height, int buttonType);

	int getEntryHeight(){ return imageEntryHeight; };
	int getFolderHeight(){ return imageFolderHeight; };
	int getInfoHeight(){ return imageInfoHeight; };
	int getSeparatorHeight(){ return imageSeparatorHeight; };
	//int getVSeparatorHeight(){ return imageVSeparatorHeight; };

	int getArrowHeight(){ return imageArrowHeight; };
	int getArrowWidth(){ return imageArrowWidth; };
	int getSelArrowHeight(){ return imageSelArrowHeight; };
	int getSelArrowWidth(){ return imageSelArrowWidth; };

	bool getTrueTransparent(){ return (	defaultentry.trueTransparent 
										 || entry.trueTransparent
										 || selentry.trueTransparent
										 || folder.trueTransparent
										 || selfolder.trueTransparent
										 || info.trueTransparent); };

private:

	bool useAlphaMap;

	void subconfigure(string uniquetype, string grouptype, BUTTONSETTINGS &workSettings, BUTTONSETTINGS &defaultSettings);

	void parseCFG(string cfg, COLORREF &bgcolor, COLORREF &lightcolor, COLORREF &darkcolor);
	void parseBevelCFG(string cfg, int &leftBevel, int &rightBevel, int &topBevel, int &bottomBevel );
	void parseGradientCFG(string cfg, int imageType);
	void solidGradientPaint(HDC destHDC, ColorRefList workGradient, RECT r, int buttonType);
	void paintEffects(HDC destHDC, int width, int height, int button_saturation, int button_hueIntensity, COLORREF button_hueColor );
	void PrepareAlphaChannel(HBITMAP &bmp, int button_saturation, int button_hueIntensity, COLORREF button_hueColor);

	ColorRefList EntryGradient;
	ColorRefList SelEntryGradient;
	ColorRefList FolderGradient;
	ColorRefList SelFolderGradient;
	ColorRefList InfoGradient;
	ColorRefList SeparatorGradient;

	HDC memoryEntryDC;
	HBITMAP memoryEntryBitmap;

	HDC memorySelEntryDC;
	HBITMAP memorySelEntryBitmap;

	HDC memoryFolderDC;
	HBITMAP memoryFolderBitmap;

	HDC memorySelFolderDC;
	HBITMAP memorySelFolderBitmap;

	HDC memoryInfoDC;
	HBITMAP memoryInfoBitmap;

	HDC memorySeparatorDC;
	HBITMAP memorySeparatorBitmap;

	HDC memoryArrowDC;
	HBITMAP memoryArrowBitmap;

	HDC memorySelArrowDC;
	HBITMAP memorySelArrowBitmap;

	int imageEntryHeight;
	int imageEntryWidth;

	int imageSelEntryHeight;
	int imageSelEntryWidth;

	int imageFolderHeight;
	int imageFolderWidth;

	int imageSelFolderHeight;
	int imageSelFolderWidth;

	int imageInfoHeight;
	int imageInfoWidth;

	int imageSeparatorHeight;
	int imageSeparatorWidth;

	int imageArrowHeight;
	int imageArrowWidth;

	int imageSelArrowHeight;
	int imageSelArrowWidth;

	BUTTONSETTINGS defaultentry;
	BUTTONSETTINGS entry;
	BUTTONSETTINGS selentry;
	BUTTONSETTINGS folder;
	BUTTONSETTINGS selfolder;
	BUTTONSETTINGS info;
	BUTTONSETTINGS separator;		

	bool activeFolderUseFolderDefault;

	int arrowSpacing;
	bool arrowOnLeft;
	int selarrowSpacing;
	bool selarrowOnLeft;

	string entryImage;
	string selentryImage;
	string folderImage;
	string selfolderImage;
	string infoImage;

		string entrysolidcolors;
		string selentrysolidcolors;
		string foldersolidcolors;
		string selfoldersolidcolors;
		string infosolidcolors;

		string entrygradientColor;
		string selentrygradientColor;
		string foldergradientColor;

	bool useentrySolidColors;
	bool entryverticalGradient;
	int entryrepeatedGradient;
	int entrybevelSizeLeft, entrybevelSizeRight, entrybevelSizeTop, entrybevelSizeBottom;
	COLORREF entrybackgroundColor;
	COLORREF entrydarkBevelColor;
	COLORREF entrylightBevelColor;

	bool useselentrySolidColors;
	bool selentryverticalGradient;
	int selentryrepeatedGradient;
	int selentrybevelSizeLeft, selentrybevelSizeRight, selentrybevelSizeTop, selentrybevelSizeBottom;
	COLORREF selentrybackgroundColor;
	COLORREF selentrydarkBevelColor;
	COLORREF selentrylightBevelColor;

	bool usefolderSolidColors;
	bool folderverticalGradient;
	int folderrepeatedGradient;
	int folderbevelSizeLeft, folderbevelSizeRight, folderbevelSizeTop, folderbevelSizeBottom;
	COLORREF folderbackgroundColor;
	COLORREF folderdarkBevelColor;
	COLORREF folderlightBevelColor;

	bool useselfolderSolidColors;
	bool selfolderverticalGradient;
	int selfolderrepeatedGradient;
	int selfolderbevelSizeLeft, selfolderbevelSizeRight, selfolderbevelSizeTop, selfolderbevelSizeBottom;
	COLORREF selfolderbackgroundColor;
	COLORREF selfolderdarkBevelColor;
	COLORREF selfolderlightBevelColor;

	bool useinfoSolidColors;
	bool infoverticalGradient;
	int inforepeatedGradient;
	int infobevelSizeLeft, infobevelSizeRight, infobevelSizeTop, infobevelSizeBottom;
	COLORREF infobackgroundColor;
	COLORREF infodarkBevelColor;
	COLORREF infolightBevelColor;

	bool useseparatorSolidColors;
	bool separatorverticalGradient;
	int separatorrepeatedGradient;
	int separatorbevelSizeLeft, separatorbevelSizeRight, separatorbevelSizeTop, separatorbevelSizeBottom;
	COLORREF separatorbackgroundColor;
	COLORREF separatordarkBevelColor;
	COLORREF separatorlightBevelColor;
};

#endif // !defined(AFX_BUTTONTEXTURE_H__4F10FDCD_EF77_423F_8223_011704DFFED2__INCLUDED_)
