/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2005 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "StdAfx.h"
#include "ContextMenu.h"

IContextMenu2 *g_IContext2 = NULL;
IContextMenu3 *g_IContext3 = NULL;

BOOL ContextMenu(LPCSTR pszPath)
{
	LPMALLOC pMalloc;
	LPSHELLFOLDER psfFolder, psfNextFolder;
	LPITEMIDLIST pidlMain, pidlItem, pidlNextItem, *ppidl;
	LPCONTEXTMENU pContextMenu;
	CMINVOKECOMMANDINFO ici;
	ULONG ulCount, ulAttr;
	char tchPath[MAX_PATH_LENGTH];
	WCHAR wchPath[MAX_PATH_LENGTH];
	UINT nCount, nCmd;
	BOOL bResult;
	HMENU hMenu;

	// Get current cursor position & window
	POINT point;
	HWND hwnd;

	GetCursorPos(&point);
	hwnd = WindowFromPoint(point);

	//
	// Get pointers to the shell's IMalloc interface and the desktop's
	// IShellFolder interface.
	//
	bResult = FALSE;

	if (!SUCCEEDED (SHGetMalloc (&pMalloc)))
		return bResult;

	if (!SUCCEEDED (SHGetDesktopFolder (&psfFolder)))
	{
		pMalloc->Release();
		return bResult;
	}
	
	//
	// Make sure the file name is fully qualified and in Unicode format.
	//
	GetFullPathName (pszPath, sizeof (tchPath) / sizeof (char), tchPath, NULL);
	//GetFullPathName (szGotPath, sizeof (tchPath) / sizeof (char), tchPath, NULL);
	if (IsTextUnicode (tchPath, lstrlen (tchPath), NULL))
		lstrcpy ((char *) wchPath, tchPath);
	else
		MultiByteToWideChar (CP_ACP, 0, tchPath, -1, wchPath,
		                     sizeof (wchPath) / sizeof (WCHAR));

	if (SUCCEEDED (psfFolder->ParseDisplayName (hwnd, NULL, wchPath, &ulCount,
				   &pidlMain, &ulAttr)) && (pidlMain != NULL))
	{
		if (nCount = GetItemCount (pidlMain))
		{ 
			// nCount must be > 0
			//
			// Initialize psfFolder with a pointer to the IShellFolder
			// interface of the folder that contains the item whose context
			// menu we're after, and initialize pidlItem with a pointer to
			// the item's item ID. If nCount > 1, this requires us to walk
			// the list of item IDs stored in pidlMain and bind to each
			// subfolder referenced in the list.
			//
			pidlItem = pidlMain;

			while (--nCount)
			{
				//
				// Create a 1-item item ID list for the next item in pidlMain.
				//
				pidlNextItem = DuplicateItem (pMalloc, pidlItem);
				if (pidlNextItem == NULL)
				{
					pMalloc->Free (pidlMain);
					psfFolder->Release();
					pMalloc->Release();
					return bResult;
				}

				//
				// Bind to the folder specified in the new item ID list.
				//
				if (!SUCCEEDED (psfFolder->BindToObject (pidlNextItem, NULL, IID_IShellFolder, (void**)&psfNextFolder)))
				{
					pMalloc->Free(pidlNextItem);
					pMalloc->Free (pidlMain);
					psfFolder->Release();
					pMalloc->Release();
					return bResult;
				}

				//
				// Release the IShellFolder pointer to the parent folder
				// and set psfFolder equal to the IShellFolder pointer for
				// the current folder.
				//
				psfFolder->Release();
				psfFolder = psfNextFolder;

				//
				// Release the storage for the 1-item item ID list we created
				// just a moment ago and initialize pidlItem so that it points
				// to the next item in pidlMain.
				//
				pMalloc->Free(pidlNextItem);
				pidlItem = GetNextItem (pidlItem);
			}
		}
	}
	
	if (pidlItem != NULL)
	{
		//
		// Get a pointer to the item's IContextMenu interface and call
		// IContextMenu::QueryContextMenu to initialize a context menu.
		//
		int iMenuType = 1;
		g_IContext2 = NULL;
		g_IContext3 = NULL;
		ppidl = &pidlItem;
		if (SUCCEEDED (psfFolder->GetUIObjectOf (hwnd, 1, (const ITEMIDLIST**)ppidl, (const GUID)IID_IContextMenu, NULL, (void**)&pContextMenu)))
		{
			// Attempt to "upgrade" to IContextMenu3/2
			LPCONTEXTMENU pCMX = NULL;
			if(SUCCEEDED(pContextMenu->QueryInterface(IID_IContextMenu3, (void**)&pCMX)))
			{
				pContextMenu->Release();
				pContextMenu = pCMX;
				iMenuType = 3;
			}
			else if(SUCCEEDED(pContextMenu->QueryInterface(IID_IContextMenu2, (void**)&pCMX)))
			{
				pContextMenu->Release();
				pContextMenu = pCMX;
				iMenuType = 2;
			}

			hMenu = CreatePopupMenu ();
			if (SUCCEEDED (pContextMenu->QueryContextMenu (
			                   hMenu, 0, 1, 0x7FFF, CMF_NORMAL)))
			{	
				// Subclass if it's version 2 or 3
				WNDPROC OldWndProc = NULL;
				if(iMenuType > 1)
				{
					OldWndProc = (WNDPROC)SetWindowLong(hwnd, GWL_WNDPROC, (DWORD)HookWndProc);
					SetProp(hwnd, TEXT("OldWndProc"), OldWndProc);
					if(iMenuType == 2)
					{
						g_IContext2 = (LPCONTEXTMENU2)pContextMenu;
					}
					else
					{
						g_IContext3 = (LPCONTEXTMENU3)pContextMenu;
					}
				}

				//
				// Display the context menu.
				//
				nCmd = TrackPopupMenu (hMenu, TPM_LEFTALIGN |
				                       TPM_LEFTBUTTON | TPM_RIGHTBUTTON | TPM_RETURNCMD,
				                       point.x, point.y, 0, hwnd, NULL);

				// Unsubclass
				if(OldWndProc) {
					SetWindowLong(hwnd, GWL_WNDPROC, (DWORD)OldWndProc);
					RemoveProp(hwnd, TEXT("OldWndProc"));
				}

				//
				// If a command was selected from the menu, execute it.
				//
				if (nCmd)
				{		
					ici.cbSize = sizeof (CMINVOKECOMMANDINFO);
					ici.fMask = SEE_MASK_FLAG_NO_UI;
					ici.hwnd = hwnd;
					ici.lpVerb = MAKEINTRESOURCE (nCmd - 1);
					ici.lpParameters = NULL;
					ici.lpDirectory = NULL;
					ici.nShow = SW_SHOWNORMAL;
					ici.dwHotKey = 0;
					ici.hIcon = NULL;

					if (SUCCEEDED (pContextMenu->InvokeCommand (&ici)))
							bResult = TRUE;	
				}
			}
			DestroyMenu (hMenu);
			pContextMenu->Release();
			g_IContext2 = NULL;
			g_IContext3 = NULL;
		}
	}

	if (pidlMain != NULL)
		pMalloc->Free(pidlMain);

	//
	// Clean up and return.
	//
	psfFolder->Release();
	pMalloc->Release();

	return bResult;
}

LRESULT CALLBACK HookWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{ 
	case WM_MENUCHAR:	// only supported by IContextMenu3
		if(g_IContext3)
		{
			LRESULT lResult = 0;
			g_IContext3->HandleMenuMsg2(message, wParam, lParam, &lResult);
			return lResult;
		}
		break;

	case WM_DRAWITEM:
	case WM_MEASUREITEM:
		if(wParam)
		{
			break; // if wParam != 0 then the message is not menu-related
		}
  
	case WM_INITMENUPOPUP:
		if(g_IContext2)
		{
			g_IContext2->HandleMenuMsg(message, wParam, lParam);
		}
		else
		{	// version 3
			g_IContext3->HandleMenuMsg(message, wParam, lParam);
		}
		return (message == WM_INITMENUPOPUP ? 0 : TRUE); // inform caller that we handled WM_INITPOPUPMENU by ourself
		break;

	default:
		break;
	}

	// call original WndProc of window to prevent undefined bevhaviour of window
	return ::CallWindowProc((WNDPROC)GetProp(hWnd, TEXT("OldWndProc")), hWnd, message, wParam, lParam);
}

UINT GetItemCount (LPITEMIDLIST pidl)
{
	USHORT nLen;
	UINT nCount;
	nCount = 0;
	while ((nLen = pidl->mkid.cb) != 0)
	{
		pidl = GetNextItem (pidl);
		nCount++;
	}
	return nCount;
}

LPITEMIDLIST GetNextItem (LPITEMIDLIST pidl)
{
	USHORT nLen;

	if ((nLen = pidl->mkid.cb) == 0)
		return NULL;

	return (LPITEMIDLIST) (((LPBYTE) pidl) + nLen);
}

LPITEMIDLIST DuplicateItem (LPMALLOC pMalloc, LPITEMIDLIST pidl)
{
	USHORT nLen;
	LPITEMIDLIST pidlNew;

	nLen = pidl->mkid.cb;
	if (nLen == 0)
		return NULL;

	pidlNew = (LPITEMIDLIST) pMalloc->Alloc (nLen + sizeof (USHORT));
	if (pidlNew == NULL)
		return NULL;

	CopyMemory (pidlNew, pidl, nLen);
	*((USHORT*) (((LPBYTE) pidlNew) + nLen)) = 0;

	return pidlNew;
}
