/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2005 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "Popup.h"

#include "VerInfo.h"

#include "PopupTexture.h"
#include "ButtonTexture.h"
#include "ButtonIcon.h"
#include "ButtonFont.h"

PopupTexture *defaultpopuptexture = NULL;
ButtonTexture *defaultbuttontexture = NULL;
ButtonIcon *defaultbuttonicon = NULL;
ButtonFont *defaultbuttonfont = NULL;

//DropShadow under XP
#ifndef CS_DROPSHADOW
	#define CS_DROPSHADOW 0x00020000
#endif

HINSTANCE hInstance;
HWND messageHandler;

bool compatibilitymode = GetxRCBool("xPopup", "CompatibilityMode");
bool cacheAllPopupsAtStartup = GetxRCBool("xPopup", "CacheAllPopupsAtStartup");

UINT (WINAPI *_PrivateExtractIcons)(LPCTSTR,int,int,int,HICON*,UINT*,UINT,UINT) = NULL;

PopupList popupList;
PopupList visiblepopupList;

int lsMessages[] = {
	LM_GETREVID,
	LM_REFRESH,
	0
};

int screenX;
int	screenY;

int infoExport = 1;

bool internaltoggle = false;

int bitsperpixel = atoi( QueryPixelDepth().c_str() );

string dragndropSource = "-1";
string dragndropDestination = "";

void StartUp();
void PreParseStepRC();

void BangRescanPopup(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangPopup(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangDynamicPopup(HWND hwndCaller, const char *pszName, const char *pszArguments);

void BangTasks(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangControlPanel(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangMyComputer(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangNetwork(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangPrinters(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangRecentDocuments(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangRecycleBin(HWND hwndCaller, const char *pszName, const char *pszArguments);

void BangReposition(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangPin(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangUnPin(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangTogglePin(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangToggleShade(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangToggle(HWND hwndCaller, const char *pszName, const char *pszArguments);

LRESULT WINAPI MessageHandlerProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
		case LM_GETREVID:
		{
			UINT uLength;
			StringCchPrintf((char*)lParam, 64, "%s.dll: %s", V_NAME, V_VERSION);
			
			if (SUCCEEDED(StringCchLength((char*)lParam, 64, &uLength)))
				return uLength;

			lParam = NULL;
			return 0;
		}

		case LM_REFRESH:
		{
			if (compatibilitymode)
				LSExecute(NULL, "!popuprescan", SW_SHOWNORMAL);
			else
				LSExecute(NULL, "!xpopuprescan", SW_SHOWNORMAL);

			return 0;
		}

		case WM_TIMER:
		{
			KillTimer(messageHandler, NULL);
			StartUp();
		}
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

// Why are we advoiding the crt startup code? The start up code is not large
// enough to cause bloat (except possibly a long system call, and I suspect that
// would need done eventually anyway) (default _DllMainCRTStartup is in
// (MSVCDIR\crt\src\dllcrt0.c) The elimination of the call to _CRT_INIT is 
// also causing warnings with MSVC7
BOOL __stdcall _DllMainCRTStartup(HINSTANCE hInst, DWORD fdwReason, LPVOID lpvRes)
{
	// We don't need thread notifications for what we're doing.  Thus, get
	// rid of them, thereby eliminating some of the overhead of this DLL
	DisableThreadLibraryCalls(hInst);
	return TRUE;
}

int initModuleEx(HWND hParent, HINSTANCE hInstance, const char *lsPath)
{
	WNDCLASSEX wc;

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS | CS_DBLCLKS;
	wc.lpfnWndProc = Popup::windowProcedure;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = sizeof(Popup *);
	wc.hInstance = hInstance;
	wc.hbrBackground = 0;
	wc.hCursor = LoadCursor(0, IDC_ARROW);
	wc.hIcon = 0;
	wc.lpszMenuName = 0;
	wc.lpszClassName = "xPopup";
	wc.hIconSm = 0;

	RegisterClassEx(&wc);

	if(IsOS(OS_XP))
	{
		wc.cbSize = sizeof(WNDCLASSEX);
		wc.style = CS_GLOBALCLASS | CS_DBLCLKS | CS_DROPSHADOW;
		wc.lpfnWndProc = Popup::windowProcedure;
		wc.cbClsExtra = 0;
		wc.cbWndExtra = sizeof(Popup *);
		wc.hInstance = hInstance;
		wc.hbrBackground = 0;
		wc.hCursor = LoadCursor(0, IDC_ARROW);
		wc.hIcon = 0;
		wc.lpszMenuName = 0;
		wc.lpszClassName = "xPopupXP";
		wc.hIconSm = 0;

		RegisterClassEx(&wc);
	}

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS;
	wc.lpfnWndProc = MessageHandlerProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hbrBackground = 0;
	wc.hCursor = 0;
	wc.hIcon = 0;
	wc.lpszMenuName = 0;
	wc.lpszClassName = "xPopupMessageHandler";
	wc.hIconSm = 0;

	RegisterClassEx(&wc);

	messageHandler = CreateWindowEx(WS_EX_TOOLWINDOW,
		"xPopupMessageHandler",
		0,
		WS_POPUP,
		0, 0, 0, 0, 
		0,
		0,
		hInstance,
		0);

	if (!messageHandler)
		return 1;
	
	SendMessage(GetLitestepWnd(),
		LM_REGISTERMESSAGE,
		(WPARAM) messageHandler,
		(LPARAM) lsMessages);

	::hInstance = hInstance;

	if (compatibilitymode)
	{
		AddBangCommandEx("!Popup", BangPopup);
		AddBangCommandEx("!PopupDynamicFolder", BangDynamicPopup);
		AddBangCommandEx("!PopupRescan", BangRescanPopup);

		AddBangCommandEx("!PopupTasks", BangTasks);
		AddBangCommandEx("!PopupControlPanel", BangControlPanel);
        AddBangCommandEx("!PopupMyComputer", BangMyComputer);
        AddBangCommandEx("!PopupNetwork", BangNetwork);
        AddBangCommandEx("!PopupPrinters", BangPrinters);
        AddBangCommandEx("!PopupRecentDocuments", BangRecentDocuments);
        AddBangCommandEx("!PopupRecycleBin", BangRecycleBin);
	}
	else
	{
		AddBangCommandEx("!xPopup", BangPopup);
		AddBangCommandEx("!xPopupDynamicFolder", BangDynamicPopup);
		AddBangCommandEx("!xPopupRescan", BangRescanPopup);

		AddBangCommandEx("!xPopupTasks", BangTasks);
		AddBangCommandEx("!xPopupControlPanel", BangControlPanel);
        AddBangCommandEx("!xPopupMyComputer", BangMyComputer);
        AddBangCommandEx("!xPopupNetwork", BangNetwork);
        AddBangCommandEx("!xPopupPrinters", BangPrinters);
        AddBangCommandEx("!xPopupRecentDocuments", BangRecentDocuments);
        AddBangCommandEx("!xPopupRecycleBin", BangRecycleBin);
	}

	if ( !IsOS( OS_95 ) && !IsOS( OS_NT4 ) )
	{
		screenX = GetSystemMetrics(SM_CXVIRTUALSCREEN);
		screenY = GetSystemMetrics(SM_CYVIRTUALSCREEN);
	}
	else
	{
		screenX = GetSystemMetrics(SM_CXSCREEN);
		screenY = GetSystemMetrics(SM_CYSCREEN);
	}

	if (!_PrivateExtractIcons)
	{
		// Bind the icon extraction function
		_PrivateExtractIcons = (UINT (WINAPI *)(LPCTSTR,int,int,int,HICON*,UINT*,UINT,UINT)) GetProcAddress(GetModuleHandle("USER32.DLL"), "PrivateExtractIconsA");
	}

	//Create "xPopup" Default Painting Classes
	defaultpopuptexture = new PopupTexture();
	defaultpopuptexture->configure("xpopup", "");
	
	defaultbuttontexture = new ButtonTexture();
	defaultbuttontexture->configure("xpopup", "");
	
	defaultbuttonicon = new ButtonIcon();
	defaultbuttonicon->configure("xpopup", "");
	
	defaultbuttonfont = new ButtonFont();
	defaultbuttonfont->configure("xpopup", "");

	//Parse Step
	if ( cacheAllPopupsAtStartup )
		SetTimer(messageHandler, NULL, 1000, 0);
	else
		StartUp();

	return 0;
}

void StartUp()
{
	PreParseStepRC();

	StringList pinnedList;
	GetxRCList( "*xPopupStartPinned", pinnedList );

	char szFirst[MAX_PATH_LENGTH] = { 0 };
	char szSecond[MAX_PATH_LENGTH] = { 0 };
	char szThird[MAX_PATH_LENGTH] = { 0 };
	char szFourth[MAX_PATH_LENGTH] = { 0 };

	char *buffer[] = { szFirst, szSecond , szThird, szFourth };

	int x = 0;
	int y = 0;
	for (StringListIterator sit = pinnedList.begin(); sit != pinnedList.end(); ++sit)
	{
		LCTokenize((*sit).c_str(), buffer, 4, NULL);

		x = xParseCoord(szSecond, 0, screenX);
		y = xParseCoord(szThird, 0, screenY);
		for(PopupListIterator plit = popupList.begin(); plit != popupList.end(); plit++)
		{
			if (stricmp((*plit)->BangToCall.c_str(), szFirst) == NULL)
			{
				if (stricmp(szFourth, "topright") == NULL)
				{
					if (x-(*plit)->getWidth() < 0)
						x = 0;
					else
						x = x-(*plit)->getWidth();

					if (y+(*plit)->getHeight() > screenY)
						y = screenY-(*plit)->getHeight();
				}
				else if(stricmp(szFourth, "bottomleft") == NULL)
				{
					if (x+(*plit)->getWidth() > screenX)
						x = screenX-(*plit)->getWidth();

					if (y-(*plit)->getHeight() < 0)
						y = 0;
					else
						y = y-(*plit)->getHeight();
				}
				else if(stricmp(szFourth, "bottomright") == NULL)
				{
					if (x-(*plit)->getWidth() < 0)
						x = 0;
					else
						x = x-(*plit)->getWidth();

					if (y-(*plit)->getHeight() < 0)
						y = 0;
					else
						y = y-(*plit)->getHeight();
				}
				// TopLeft is Default
				else
				{
					if (x+(*plit)->getWidth() > screenX)
						x = screenX-(*plit)->getWidth();
					if (y+(*plit)->getHeight() > screenY)
						y = screenY-(*plit)->getHeight();
				}

				(*plit)->reposition(3, x, y, (*plit)->getWidth(), (*plit)->getHeight());
				(*plit)->PinnedMode = true;
				if ( (*plit)->pinnedPopupNotOnTop )
					(*plit)->setAlwaysOnTop(false);
				else
					(*plit)->setAlwaysOnTop(true);
				(*plit)->show();
				break;
			}
		}		
	}
	pinnedList.clear();
}

void PreParseStepRC()
{
	StringList settingsList;

	if (compatibilitymode)
		GetxRCList( "*Popup", settingsList );
	else
		GetxRCList( "*xPopup", settingsList );

	char szFirst[MAX_PATH_LENGTH] = { 0 };
	char szSecond[MAX_PATH_LENGTH] = { 0 };
	char szThird[MAX_PATH_LENGTH] = { 0 };
	char szFourth[MAX_PATH_LENGTH] = { 0 };

	char *buffer[] = { szFirst, szSecond , szThird, szFourth };

	Popup* addsettingPopup = NULL;
	bool input = false;
	StringList sendToPopup;
	string check;
	for (StringListIterator sit = settingsList.begin(); sit != settingsList.end(); ++sit)
	{
		int nTokens = LCTokenize((*sit).c_str(), buffer, 3, szFourth);

		if(nTokens >= 2 && (stricmp(szFirst, "!New") == NULL || stricmp(szSecond, "!New") == NULL))
		{	
			//Search for Popup Start with !New and Create New Popup
			Popup *popup = NULL;
			if (stricmp(szFirst, "!New") == NULL)
			{
				check = szThird;
				if (check.empty())
					popup = new Popup("xpopup", "", defaultpopuptexture, defaultbuttontexture, defaultbuttonicon, defaultbuttonfont);
				else
					popup = new Popup(szThird, "");
				popup->BangToCall = szSecond;
				AddBangCommandEx(szSecond, BangPopup);

				//Reposition Bang
				string bang = szSecond;
				bang.append("Reposition");
				AddBangCommandEx(bang.c_str(), BangReposition);

				//Pin Bang == Show
				bang = szSecond;
				bang.append("Pin");
				AddBangCommandEx(bang.c_str(), BangPin);

				//UnPin Bang == Hide
				bang = szSecond;
				bang.append("UnPin");
				AddBangCommandEx(bang.c_str(), BangUnPin);

				//Toggle PinnedMode == Toggle
				bang = szSecond;
				bang.append("TogglePin");
				AddBangCommandEx(bang.c_str(), BangTogglePin);

				//Toggle ShadeMode == Toggle
				bang = szSecond;
				bang.append("ToggleShade");
				AddBangCommandEx(bang.c_str(), BangToggleShade);

				//Toggle Visibility == Toggle
				bang = szSecond;
				bang.append("Toggle");
				AddBangCommandEx(bang.c_str(), BangToggle);
			}
			else if (stricmp(szSecond, "!New") == NULL)
			{
				check = szFourth;
				if (check.empty())
					popup = new Popup("xpopup", szFirst, defaultpopuptexture, defaultbuttontexture, defaultbuttonicon, defaultbuttonfont);
				else
					popup = new Popup(szFourth, szFirst);
				popup->BangToCall = szThird;
				AddBangCommandEx(szThird, BangPopup);

				//Reposition Bang
				string bang = szThird;
				bang.append("Reposition");
				AddBangCommandEx(bang.c_str(), BangReposition);

				//Pin Bang == Show
				bang = szThird;
				bang.append("Pin");
				AddBangCommandEx(bang.c_str(), BangPin);

				//UnPin Bang == Hide
				bang = szThird;
				bang.append("UnPin");
				AddBangCommandEx(bang.c_str(), BangUnPin);

				//Toggle PinnedMode == Toggle
				bang = szThird;
				bang.append("TogglePin");
				AddBangCommandEx(bang.c_str(), BangTogglePin);

				//Toggle ShadeMode == Toggle
				bang = szThird;
				bang.append("ToggleShade");
				AddBangCommandEx(bang.c_str(), BangToggleShade);

				//Toggle Visibility == Toggle
				bang = szThird;
				bang.append("Toggle");
				AddBangCommandEx(bang.c_str(), BangToggle);
			}
		
			if (popup)
			{
				popupList.insert(popupList.end(), popup);
						
				addsettingPopup = popup;

				//Allow following settings to be tracked
				input = true;
			}
		}
		else if (nTokens >= 1 && (stricmp(szFirst, "~New") == NULL || stricmp(szSecond, "~New") == NULL))
		{
			//Close Popup Definition, give Settings and Load
			if (addsettingPopup)
			{
				addsettingPopup->load(hInstance);
				addsettingPopup->popupcontentList = sendToPopup;
				addsettingPopup->ParseStepRC();
			}
			addsettingPopup = NULL;
			input = false;
			sendToPopup.clear();
		}
		else if (input)
		{
			//Add to specific Popup Settings list
			if (strstr(szFirst, ".icon=") == NULL && strstr(szFirst, ".Icon=") == NULL && strstr(szFirst, ".ICON=") == NULL)
			{
				sendToPopup.insert(sendToPopup.end(), (".icon=\".noexist\" " + *sit));
			}
			else
			{
				sendToPopup.insert(sendToPopup.end(), *sit);
			}
		}
	}

	sendToPopup.clear();
}

void quitModule(HINSTANCE hInstance)
{
	visiblepopupList.clear();

	for(PopupListIterator plit = popupList.begin(); plit != popupList.end(); plit++)
	{
		(*plit)->hide();
		delete *plit;
	}

	popupList.clear();

	if (compatibilitymode)
	{
		RemoveBangCommand("!Popup");
		RemoveBangCommand("!PopupDynamicFolder");
		RemoveBangCommand("!PopupRescan");

		RemoveBangCommand("!PopupTasks");
		RemoveBangCommand("!PopupControlPanel");
        RemoveBangCommand("!PopupMyComputer");
        RemoveBangCommand("!PopupNetwork");
        RemoveBangCommand("!PopupPrinters");
        RemoveBangCommand("!PopupRecentDocuments");
        RemoveBangCommand("!PopupRecycleBin");
	}
	else
	{
		RemoveBangCommand("!xPopup");
		RemoveBangCommand("!xPopupDynamicFolder");
		RemoveBangCommand("!xPopupRescan");

		RemoveBangCommand("!xPopupTasks");
		RemoveBangCommand("!xPopupControlPanel");
        RemoveBangCommand("!xPopupMyComputer");
        RemoveBangCommand("!xPopupNetwork");
        RemoveBangCommand("!xPopupPrinters");
        RemoveBangCommand("!xPopupRecentDocuments");
        RemoveBangCommand("!xPopupRecycleBin");
	}

	if (defaultpopuptexture)
		delete defaultpopuptexture;
	if (defaultbuttontexture)
		delete defaultbuttontexture;
	if (defaultbuttonicon)
		delete defaultbuttonicon;
	if (defaultbuttonfont)
		delete defaultbuttonfont;

	SendMessage(GetLitestepWnd(),
		LM_UNREGISTERMESSAGE,
		(WPARAM) messageHandler,
		(LPARAM) lsMessages);

	DestroyWindow(messageHandler);

	UnregisterClass("xPopup", hInstance);
	if(IsOS(OS_XP))
		UnregisterClass("xPopupXP", hInstance);
	UnregisterClass("xPopupMessageHandler", hInstance);
}

void BangPopup(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	char x[32] = {0};
	char y[32] = {0};
	char anchor[64] = {0};
	char *posbuffers[] = {x, y, anchor};
	
	int numTokens = LCTokenize( pszArguments, posbuffers, 3, NULL);

	POINT cursorPos;
	GetCursorPos(&cursorPos);
	if (numTokens >= 2)
	{
		cursorPos.x = xParseCoord(x, cursorPos.x, screenX);
		cursorPos.y = xParseCoord(y, cursorPos.y, screenY);
	}

	bool oneshown = false;
	for(PopupListIterator plit = popupList.begin(); plit != popupList.end(); plit++)
	{
		if ( !oneshown && (stricmp( (*plit)->BangToCall.c_str(), pszName) == NULL || (compatibilitymode && stricmp( pszName, "!popup") == NULL) || (!compatibilitymode && stricmp( pszName, "!xpopup") == NULL)))
		{
			//Hide, if already visible
			if ( (*plit)->isVisible() && internaltoggle )
			{
				(*plit)->hide();
				oneshown = true;
				continue;
			}

			if ( (*plit)->topLevelSpecialPopup == 1 )
				(*plit)->ScanSubMenu((*plit)->getRootPath(), -2);
			else if ( (*plit)->topLevelSpecialPopup == 2 )
				(*plit)->ScanTasks();

			if ( stricmp(anchor, "topleft") != NULL && stricmp(anchor, "topright") != NULL
					&& stricmp(anchor, "bottomleft") != NULL && stricmp(anchor, "bottomright") != NULL )
			{
				if (cursorPos.y+(*plit)->getHeight() > screenY)
					strcpy(anchor, "bottom");
				else
					strcpy(anchor, "top");
					
				if (cursorPos.x+(*plit)->getWidth() > screenX)
					strcat(anchor, "right");
				else
					strcat(anchor, "left");
			}

			// TopLeft is Default
			if(stricmp(anchor, "topleft") == NULL)
			{
				if (cursorPos.x+(*plit)->getWidth() > screenX)
					cursorPos.x = screenX-(*plit)->getWidth();
				if (cursorPos.y+(*plit)->getHeight() > screenY)
					cursorPos.y = screenY-(*plit)->getHeight();
			}
			else if (stricmp(anchor, "topright") == NULL)
			{
				if (cursorPos.x-(*plit)->getWidth() < 0)
					cursorPos.x = 0;
				else
					cursorPos.x = cursorPos.x-(*plit)->getWidth();

				if (cursorPos.y+(*plit)->getHeight() > screenY)
					cursorPos.y = screenY-(*plit)->getHeight();
			}
			else if(stricmp(anchor, "bottomleft") == NULL)
			{
				if (cursorPos.x+(*plit)->getWidth() > screenX)
					cursorPos.x = screenX-(*plit)->getWidth();

				if (cursorPos.y-(*plit)->getHeight() < 0)
					cursorPos.y = 0;
				else
					cursorPos.y = cursorPos.y-(*plit)->getHeight();
			}
			else if(stricmp(anchor, "bottomright") == NULL)
			{
				if (cursorPos.x-(*plit)->getWidth() < 0)
					cursorPos.x = 0;
				else
					cursorPos.x = cursorPos.x-(*plit)->getWidth();

				if (cursorPos.y-(*plit)->getHeight() < 0)
					cursorPos.y = 0;
				else
					cursorPos.y = cursorPos.y-(*plit)->getHeight();
			}

			KillTimer((*plit)->gethWnd(), 1);
			(*plit)->reposition(3, cursorPos.x, cursorPos.y, (*plit)->getWidth(), (*plit)->getHeight());
			if ((*plit)->PinnedMode)
			{
				(*plit)->PinnedMode = false;
				(*plit)->setAlwaysOnTop(true);
			}
	
			(*plit)->show();
			
			oneshown = true;
		}
		else if ( (*plit)->isVisible() )
			(*plit)->hide();
	}

	internaltoggle = false;
}

void BangDynamicPopup(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	char path[256] = {0};
	char x[32] = {0};
	char y[32] = {0};
	char anchor[64] = {0};
	char *posbuffers[] = {path, x, y, anchor};
	
	int numTokens = LCTokenize( pszArguments, posbuffers, 4, NULL);

	POINT cursorPos;
	GetCursorPos(&cursorPos);

	if ( stricmp(pszName, "controlpanel") == NULL || stricmp(pszName, "mycomputer") == NULL ||
			stricmp(pszName, "network") == NULL	|| stricmp(pszName, "printers") == NULL ||
			stricmp(pszName, "recentdocuments") == NULL || stricmp(pszName, "recyclebin") == NULL ||
			stricmp(pszName, "tasks") == NULL )
	{
		strcpy(anchor, y);
		strcpy(y, x);
		strcpy(x, path);

		if (numTokens >= 2)
		{
			cursorPos.x = xParseCoord(x, cursorPos.x, screenX);
			cursorPos.y = xParseCoord(y, cursorPos.y, screenY);
		}
	}
	else
	{
		if (numTokens >= 3)
		{
			cursorPos.x = xParseCoord(x, cursorPos.x, screenX);
			cursorPos.y = xParseCoord(y, cursorPos.y, screenY);
		}
	}

	Popup* popup = NULL;
	bool dummyCreated = false;
	for(PopupListIterator plit = popupList.begin(); plit != popupList.end(); plit++)
	{
		if ( stricmp( (*plit)->BangToCall.c_str(), "dynamicdummy") == NULL )
		{
			dummyCreated = true;
			popup = *plit;
			break;
		}
	}
	if (!dummyCreated)
	{
		popup = new Popup("xpopup", "", defaultpopuptexture, defaultbuttontexture, defaultbuttonicon, defaultbuttonfont);

		popupList.insert(popupList.end(), popup);
		
		popup->BangToCall = "dynamicdummy";
					
		popup->load(hInstance);
	}
	
	if (popup != NULL)
	{
		if ( stricmp(pszName, "ControlPanel") == NULL || stricmp(pszName, "mycomputer") == NULL ||
				stricmp(pszName, "network") == NULL	|| stricmp(pszName, "printers") == NULL ||
				stricmp(pszName, "recentdocuments") == NULL || stricmp(pszName, "recyclebin") == NULL ||
				stricmp(pszName, "tasks") == NULL )
		{
			int clsid = 0;

			if ( stricmp(pszName, "controlpanel") == NULL )
			{
				clsid = CSIDL_CONTROLS;
				popup->captionname = "ControlPanel";
			}
			else if ( stricmp(pszName, "mycomputer") == NULL )
			{
				clsid = CSIDL_DRIVES;
				popup->captionname = "MyComputer";
			}
			else if ( stricmp(pszName, "network") == NULL )
			{
				clsid = CSIDL_NETWORK;
				popup->captionname = "Network";
			}
			else if ( stricmp(pszName, "printers") == NULL )
			{
				clsid = CSIDL_PRINTERS;
				popup->captionname = "Printers";
			}
			else if ( stricmp(pszName, "recentdocuments") == NULL )
			{
				clsid = CSIDL_RECENT;
				popup->captionname = "RecentDocuments";
			}
			else if ( stricmp(pszName, "recyclebin") == NULL )
			{
				clsid = CSIDL_BITBUCKET;
				popup->captionname = "RecycleBin";
			}
			else if ( stricmp(pszName, "tasks") == NULL )
			{
				clsid = -1;
				popup->captionname = "Tasks";
			}

			if (clsid == 0)
				return;

			if ( clsid == -1 )
				popup->ScanTasks();
			else
				popup->ScanShellFolder(clsid);
		}
		else
		{
			CString easyWorker = path;
			if (easyWorker.Right(1) == '\\')
				easyWorker = easyWorker.Left(easyWorker.GetLength()-1);
			easyWorker = easyWorker.Right(easyWorker.GetLength()-easyWorker.ReverseFind('\\')-1);
			popup->captionname = (string) easyWorker;
			popup->ScanSubMenu((string) path, true);
		}

		if (stricmp(anchor, "topright") == NULL)
		{
			if (cursorPos.x-popup->getWidth() < 0)
				cursorPos.x = 0;
			else
				cursorPos.x = cursorPos.x-popup->getWidth();

			if (cursorPos.y+popup->getHeight() > screenY)
				cursorPos.y = screenY-popup->getHeight();
		}
		else if(stricmp(anchor, "bottomleft") == NULL)
		{
			if (cursorPos.x+popup->getWidth() > screenX)
				cursorPos.x = screenX-popup->getWidth();

			if (cursorPos.y-popup->getHeight() < 0)
				cursorPos.y = 0;
			else
				cursorPos.y = cursorPos.y-popup->getHeight();
		}
		else if(stricmp(anchor, "bottomright") == NULL)
		{
			if (cursorPos.x-popup->getWidth() < 0)
				cursorPos.x = 0;
			else
				cursorPos.x = cursorPos.x-popup->getWidth();

			if (cursorPos.y-popup->getHeight() < 0)
				cursorPos.y = 0;
			else
				cursorPos.y = cursorPos.y-popup->getHeight();
		}
		// TopLeft is Default
		else
		{
			if (cursorPos.x+popup->getWidth() > screenX)
				cursorPos.x = screenX-popup->getWidth();
			if (cursorPos.y+popup->getHeight() > screenY)
				cursorPos.y = screenY-popup->getHeight();
		}

		popup->reposition(1, cursorPos.x, cursorPos.y, 0, 0);
		popup->show();
	}
}

void BangTasks(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	BangDynamicPopup(hwndCaller, "Tasks", pszArguments);
}

void BangControlPanel(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	BangDynamicPopup(hwndCaller, "ControlPanel", pszArguments);
}

void BangMyComputer(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	BangDynamicPopup(hwndCaller, "MyComputer", pszArguments);
}

void BangNetwork(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	BangDynamicPopup(hwndCaller, "Network", pszArguments);
}

void BangPrinters(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	BangDynamicPopup(hwndCaller, "Printers", pszArguments);
}

void BangRecentDocuments(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	BangDynamicPopup(hwndCaller, "RecentDocuments", pszArguments);
}

void BangRecycleBin(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	BangDynamicPopup(hwndCaller, "RecycleBin", pszArguments);
}

void BangReposition(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	int x;
	int y;
	char A[16], B[16], C[16], D[16], E[32];
	char *buffers[] = {A, B, C, D, E};

	int tokens = LCTokenize(pszArguments, buffers, 5, NULL);

	if ( tokens >= 2 )
	{
		for(PopupListIterator plit = popupList.begin(); plit != popupList.end(); plit++)
		{
			if ( stricmp( ((*plit)->BangToCall + "reposition").c_str(), pszName) == NULL )
			{	
				if ( stricmp(A, "currentx") == NULL)
					x = (*plit)->getX();
				else
					x = xParseCoord(A, 0, screenX);
				if ( stricmp(B, "currenty") == NULL)
					y = (*plit)->getY();
				else
					y = xParseCoord(B, 0, screenY);

				if (tokens == 2)
				{
					(*plit)->reposition(1, x, y, 0, 0);
				}
				else if (tokens == 3)
				{
					if (stricmp(C, "topright") == NULL)
					{
						x = x-(*plit)->getWidth();
						(*plit)->reposition(1, x, y, 0, 0);
					}
					else if(stricmp(C, "bottomleft") == NULL)
					{
						y = y-(*plit)->getHeight();
						(*plit)->reposition(1, x, y, 0, 0);
					}
					else if(stricmp(C, "bottomright") == NULL)
					{
						x = x-(*plit)->getWidth();
						y = y-(*plit)->getHeight();
						(*plit)->reposition(1, x, y, 0, 0);
					}
					else
					{
						(*plit)->reposition(1, x, y, 0, 0);
					}
				}
				else if (tokens == 4)
				{
					(*plit)->reposition(3, x, y, atoi( C ), atoi( D ), 1);
				}
				else if (tokens == 5)
				{
					if (stricmp(E, "topright") == NULL)
					{
						x = x-(*plit)->getWidth();
			
						(*plit)->reposition(3, x, y, atoi( C ), atoi( D ), 1);
					}
					else if(stricmp(E, "bottomleft") == NULL)
					{
						y = y-(*plit)->getHeight();
					
						(*plit)->reposition(3, x, y, atoi( C ), atoi( D ), 1);
					}
					else if(stricmp(E, "bottomright") == NULL)
					{
						x = x-(*plit)->getWidth();
						y = y-(*plit)->getHeight();
						
						(*plit)->reposition(3, x, y, atoi( C ), atoi( D ), 1);
					}
					else
					{
						(*plit)->reposition(3, x, y, atoi( C ), atoi( D ), 1);
					}
				}

				break;
			}
		}
	}
}

void BangPin(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	for(PopupListIterator plit = popupList.begin(); plit != popupList.end(); plit++)
	{
		if ( stricmp( ((*plit)->BangToCall + "pin").c_str(), pszName) == NULL )
		{
			(*plit)->PinnedMode = true;
			if ( (*plit)->pinnedPopupNotOnTop )
				(*plit)->setAlwaysOnTop(false);
			else
				(*plit)->setAlwaysOnTop(true);
			(*plit)->show();
			break;
		}
	}
}

void BangUnPin(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	for(PopupListIterator plit = popupList.begin(); plit != popupList.end(); plit++)
	{
		if ( stricmp( ((*plit)->BangToCall + "unpin").c_str(), pszName) == NULL )
		{
			(*plit)->PinnedMode = false;
			(*plit)->setAlwaysOnTop(true);
			(*plit)->hide();
			break;
		}
	}
}

void BangTogglePin(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	for(PopupListIterator plit = popupList.begin(); plit != popupList.end(); plit++)
	{
		if ( stricmp( ((*plit)->BangToCall + "togglepin").c_str(), pszName) == NULL )
		{
			if ((*plit)->PinnedMode)
			{
				(*plit)->PinnedMode = false;
				(*plit)->setAlwaysOnTop(true);
				(*plit)->hide();
			}
			else
			{
				(*plit)->PinnedMode = true;
				if ( (*plit)->pinnedPopupNotOnTop )
					(*plit)->setAlwaysOnTop(false);
				else
					(*plit)->setAlwaysOnTop(true);
				(*plit)->show();
			}
			break;
		}
	}
}

void BangToggleShade(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	for(PopupListIterator plit = popupList.begin(); plit != popupList.end(); plit++)
	{
		if ( stricmp( ((*plit)->BangToCall + "toggleshade").c_str(), pszName) == NULL )
		{
			(*plit)->ToggleShade();

			break;
		}
	}
}

void BangToggle(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	CString convert = pszName;
	convert = convert.Left(convert.GetLength()-6);

	internaltoggle = true;

	BangPopup(hwndCaller, convert, pszArguments);
}

void BangRescanPopup(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	if ( stricmp(pszArguments, "reload") == NULL )
		LSExecute(NULL, "!reload", SW_SHOWNORMAL);

	visiblepopupList.clear();

	for(PopupListIterator plit = popupList.begin(); plit != popupList.end(); plit++)
	{
		(*plit)->hide();
		delete *plit;
	}

	popupList.clear();

	if (defaultpopuptexture)
		delete defaultpopuptexture;
	if (defaultbuttontexture)
		delete defaultbuttontexture;
	if (defaultbuttonicon)
		delete defaultbuttonicon;
	if (defaultbuttonfont)
		delete defaultbuttonfont;

	//Create xPopup Default Painting Classes
	defaultpopuptexture = new PopupTexture();
	defaultpopuptexture->configure("xpopup", "");
	
	defaultbuttontexture = new ButtonTexture();
	defaultbuttontexture->configure("xpopup", "");
	
	defaultbuttonicon = new ButtonIcon();
	defaultbuttonicon->configure("xpopup", "");
	
	defaultbuttonfont = new ButtonFont();
	defaultbuttonfont->configure("xpopup", "");

	StartUp();
}